#
# Copyright 2018 Versentiedge Inc. All Rights Reserved.
#   *** partially modified version previously published by Google Inc. ***
#   - removed unnecessary dependency on absl.flags
#   - switched platforms module to local version (not pysc2 version)
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configs for various ways to run starcraft."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

#from pysc2.run_configs import platforms
#from pysc2 import run_configs
from sc2gameLobby.runConfigs import platforms
from pysc2.run_configs import lib


################################################################################
def get(sc2_run_config=None):
  """Get the current platform OS config (else as specified)"""
  configs = {c.name(): c
             for c in lib.RunConfig.all_subclasses() if c.priority()}

  if not configs:
    raise lib.SC2LaunchError("No valid run_configs found.")

  #if FLAGS.sc2_run_config is None:  # Find the highest priority as default.
  if sc2_run_config is None:
    return max(configs.values(), key=lambda c: c.priority())()

  try:
    return configs[sc2_run_config]()
  except KeyError:
    raise lib.SC2LaunchError("Invalid run_config. Valid configs are: %s" % (
        ", ".join(sorted(configs.keys()))))





from s2clientprotocol.common_pb2 import Point2D
from s2clientprotocol.debug_pb2 import DebugCommand
from s2clientprotocol.debug_pb2 import DebugCreateUnit
from s2clientprotocol.debug_pb2 import DebugKillUnit
from s2clientprotocol.debug_pb2 import DebugSetUnitValue


################################################################################
def create(*units):
    """create this unit within the game as specified"""
    ret = []
    for unit in units: # implemented using sc2simulator.ScenarioUnit
        x, y = unit.position[:2]
        pt = Point2D(x=x, y=y)
        unit.tag = 0 # forget any tag because a new unit will be created
        new = DebugCommand(create_unit=DebugCreateUnit(
            unit_type   = unit.code,
            owner       = unit.owner,
            pos         = pt,
            quantity    = 1,
        ))
        ret.append(new)
    return ret


################################################################################
def modify(*units):
    """set the unit defined by in-game tag with desired properties"""
    ret = []
    for unit in units: # add one command for each attribute
        for attr, idx in [("energy", 1), ("life", 2), ("shields", 3)]: # see debug_pb2.UnitValue for enum declaration
            newValue = getattr(unit, attr)
            if not newValue: continue # don't bother setting something that isn't necessary
            if attr != "energy": continue
            new = DebugCommand(unit_value=DebugSetUnitValue(
                value       = 125.0,#newValue,
                unit_value  = idx,
                unit_tag    = unit.tag))
            if attr == "energy": print(new)
            ret.append(new)
    return ret


################################################################################
def remove(*tags):
    """remove the in-game units specified by tags from the game"""
    return DebugCommand(kill_unit=DebugKillUnit(tag=tags))





"""
Copyright 2018 Versentiedge LLC All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS-IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import base64
import importlib
import json
import os
import re
import requests
import subprocess
import sys
import time

from sc2players import addPlayer, getPlayer, PlayerPreGame, PlayerRecord
from sc2ladderMgmt import getLadder

from sc2gameLobby.gameConfig import Config
from sc2gameLobby import connectToServer
from sc2gameLobby import gameConstants as c
from sc2gameLobby import genericObservation as go
from sc2gameLobby import host, join
from sc2gameLobby import resultHandler as rh
from sc2gameLobby import versions


################################################################################
def exitStatement(msg, code=1):
    printMsg = msg
    if code: printMsg = "ERROR: %s"%(msg)
    print("%s%s"%(os.linesep, printMsg))
    return code


################################################################################
def badConnect(ladder, code=2):
    return exitStatement("A connection could not be made. %s may not be "
    "available or you may not be connected to the internet."%(ladder), code=code)


################################################################################
def getLaunchConfig(options):
    player = PlayerPreGame(getPlayer(options.player), selectedRace=options.race, observe=options.observe)
    ret = Config( # create game config
            expo        = c.EXPO_SELECT[options.exp],
            version     = options.version or versions.handle.mostRecent['version'],
            ladder      = getLadder(options.ladder),
            players     = [player],
            whichPlayer = player.name,
            mode        = c.types.GameModes(options.mode),
            themap      = options.map,# or selectMap(options.map),
            numObservers= options.obs,
            trust       = True,
           #slaves      = [],
            fogDisabled = options.nofog,
            stepSize    = options.step,
            opponents   = options.opponents.split(',') if options.opponents else [],
            fullscreen  = not options.windowed,
            raw         = options.raw,
            score       = options.score,
            feature     = options.feature,
            render      = options.rendered,
           #resolution
           #minimap
           #layerwidth
           #replay      = options.replay,
           #debug       = options.debug,
        )
    ret.connection # force generation of IP address and ports attributes 
    return ret


################################################################################
def run(options, cfg=None):
    if options.search or options.history:
        if not options.player: options.player = options.search
        if cfg == None:  cfg = getLaunchConfig(options)
        try: httpResp = connectToServer.ladderPlayerInfo(cfg, options.search, getMatchHistory=options.history)
        except requests.exceptions.ConnectionError:
            return badConnect(cfg.ladder)
        if not httpResp.ok: return exitStatement(httpResp.text)
        printStr = "%15s : %s"
        for playerAttrs, playerHistory in httpResp.json():
            player = PlayerRecord(source=playerAttrs)
            print(player)
            for k in ["type", "difficulty", "initCmd", "rating"]:
                print(printStr%(k, getattr(player, k)))
            if "created" in playerAttrs:    print(printStr%("created", time.strftime('%Y-%m-%d', time.localtime(player.created))))
            if playerHistory: # TODO -- do something with playerHistory, when implemented
                for h in playerH:   print(printStr%("history", h))
            print()
    elif options.nogui:
        if cfg == None:  cfg = getLaunchConfig(options)
        if False: # only display in debug/verbose mode?
            print("REQUESTED CONFIGURATION")
            cfg.display()
            print(cfg.whoAmI().initCmd)
            print()
        try:    httpResp = connectToServer.sendMatchRequest(cfg)
        except requests.exceptions.ConnectionError:
            return badConnect(cfg.ladder)
        ### cancel match request ### (would have to be issued as subprocess after match request, but before a match is assigned
            #import time
            #time.sleep(1)
            #print(connectToServer.cancelMatchRequest(cfg).text)
        if not httpResp.ok: return exitStatement(httpResp.text)
        data = httpResp.json()
        for pData in data["players"]: # if player matchup doesn't exist locally, retrieve server information and define the player
            pName = pData[0]
            try:    getPlayer(pName) # test whether player exists locally
            except ValueError: # player w/ name pName is not defined locally
                try: y = connectToServer.ladderPlayerInfo(cfg, pName) # get player info from ladder
                except requests.exceptions.ConnectionError:
                    return badConnect(cfg.ladder)
                settings = y[0][0] # settings of player[0]
                del settings["created"] # creation data is retained locally
                addPlayer(settings) # ensures that loading json + inflation succeeds
        matchCfg = Config()
        matchCfg.loadJson(data)
        print("SERVER-ASSIGNED CONFIGURATION")
        matchCfg.display()
        print()
        ### launch game; collect results ###
        thisPlayer = matchCfg.whoAmI()
        result     = None
        replayData = ""
        callback   = go.doNothing # play with human control only (default)
        if thisPlayer.initCmd: # launch the desired player appropriately
            if re.search("^\w+\.[\w\.]+$", thisPlayer.initCmd): # found a python command; extract callback and agent process/object stuff
                parts = thisPlayer.initCmd.split(".")
                moduleName = parts[0]
                try:
                    thing = importlib.import_module(moduleName)
                    for part in parts[1:]: # access callable defined by the agent
                        thing = getattr(thing, part)
                    callback = thing() # execute to acquire a list of the callback and any additional, to-be-retained objects necessary to run the agent process
                except ModuleNotFoundError as e:    return exitStatement("agent %s initialization command (%s) did not begin with a module (expected: %s). Given: %s"%(thisPlayer.name, thisPlayer.initCmd, moduleName, e))
                except AttributeError as e:         return exitStatement("invalid %s init command format (%s): %s"%(thisPlayer, thisPlayer.initCmd, e))
                except Exception as e:              return exitStatement("general failure to initialize agent %s: %s %s"%(thisPlayer.name, type(e), e))
            else: # launch separate process that manages the agent and results
                # TODO -- ensure proper port info is supplied to the subprocess
                # TODO -- ensure proper configuration info is supplied to the subprocess
                p = subprocess.Popen(thisPlayer.initCmd.split())
                p.communicate()
                msg = "Command-line bot %s finished normally."%(thisPlayer)
                if p.returncode: # otherwise rely on client to send game result to server (else server catches the non-reporting offender)
                    msg = "Command-line bot %s crashed (%d)."%(
                        thisPlayer, p.returncode)
                    try: # while non-python agent manages communication including match completion reporting to ladder server, some crash events can be reported by this process
                        httpResp = connectToServer.reportMatchCompletion(
                            matchCfg, result, replayData)
                        if not httpResp.ok: msg += " %s!"%(httpResp.text)
                    except requests.exceptions.ConnectionError: # process can't report game result if it crashes
                        msg += " lost prior connection to %s!"%(cfg.ladder)
                return exitStatement(msg, code=p.returncode)
        try:
            if matchCfg.host: # host contains details of the host to connect to
                  func = join # join game on host when host details are provided
            else: func = host # no value means this machine is hosting
            result,replayData = func(matchCfg, agentCallBack=callback,
                                     scenario=options.scenario)
        except c.TimeoutExceeded as e: # match failed to launch
            print(e)
            result = rh.launchFailure(matchCfg) # report UNDECIDED results
        finally:
            if hasattr(callback, "shutdown"):
                callback.shutdown() # close dependencies appropriately
            callback = None
        ### simulate sending match results ###
            #results = []
            #from numpy.random import choice
            #from time import sleep
            #import json
            #print("*"*80)
            #for i,p in enumerate(matchCfg.players):
            #    playerName = p.name
            #    if i%2: results.append( [playerName, int(not results[-1][1])] )
            #    else:   results.append( [playerName, int(choice([0, 1]))] )
            #sleep(1.5)
            #results = dict(results)
            #print(json.dumps(results, indent=4))
            #try:
            #    httpResp = connectToServer.reportMatchCompletion(matchCfg, results, "")
            #    if not httpResp.ok: return exitStatement(httpResp.text)
            #except requests.exceptions.ConnectionError:
            #    return badConnect(cfg.ladder)
            #print(httpResp.json())
        ### send actual results ###
        replaySize = len(replayData) if replayData else 0
        if replaySize:
            print("FINAL RESULT: (%d)"%(replaySize))
            print(json.dumps(result, indent=4))
            if options.savereplay: # save the replay if the option is specified
                dirName = os.path.dirname(options.savereplay)
                if dirName and not os.path.isdir(dirName):
                    os.makedirs(dirName)
                with open(options.savereplay, "wb") as f:
                    f.write(replayData)
        if result != None: # regular result is expected to be reported by each player
            try: # ensure replay data can be transmitted over http
                replayData = base64.encodestring(replayData).decode() # convert raw bytes into str
                httpResp = connectToServer.reportMatchCompletion(
                    matchCfg, result, replayData)
                if not httpResp.ok: return exitStatement(httpResp.text)
                if httpResp.text and "invalid" in httpResp.text:
                    return ValueError
            except requests.exceptions.ConnectionError:
                return badConnect(cfg.ladder)
            except Exception as e: return exitStatement("%s: %s"%(type(e), e))
            if replaySize: print(httpResp.json()) # also display effective rating changes
    elif options.add:       versions.addNew(*options.add.split(','))
    elif options.update:
        keys = [
            "label",
            "version",
            "base-version",
            "data-hash",
            "fixed-hash",
            "replay-hash",
        ]
        data = {}
        for k,v in zip(keys, options.update.split(',')):
            data[k] = v
        versions.handle.update(data)
        versions.handle.save()
    elif options.versions: # simply display the jsonData reformatted
        for v,record in sorted(versions.handle.ALL_VERS_DATA.items()):
            print(v)
            for k,v in sorted(record.items()):
                print("%15s: %s"%(k,v))
            print()
    else: # without an action explicitly specified, launch the GUI with selected options
        if cfg == None:  cfg = getLaunchConfig(options)
        cfg.display()
        # TODO -- launch GUI that manages communication to server
        print("ERROR: GUI mode is not yet implemented.")





from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

from s2clientprotocol import sc2api_pb2 as sc_pb

from six import iteritems # python 2/3 compatibility

import glob
import json
import os
import portpicker
import random
import re
import time

from sc2common    import types
from sc2gameLobby import dateFormat
from sc2gameLobby import gameConstants as c
from sc2gameLobby import ipAddresses
#import mapData
from sc2gameLobby import runConfigs
from sc2gameLobby import versions

from sc2players import getPlayer, buildPlayer, PlayerRecord, PlayerPreGame
from sc2players import constants as playerConstants
from sc2ladderMgmt.ladders import Ladder
from sc2maptool.functions import selectMap
from sc2maptool.mapRecord import MapRecord

"""
establish a configuration file containing game information such that other
processes can identify what any other established configuration already is
without necessitating inter-process communication for this information.
"""


################################################################################
def activeConfigs():
    """identify all configurations which are currently active"""
    #return [os.path.abspath(item) for item in os.listdir(c.FOLDER_ACTIVE_CONFIGS)]
    return glob.glob( os.path.join(c.FOLDER_ACTIVE_CONFIGS, "*.json") )


################################################################################
def clearConfigs():
    """remove all active configurations, if any"""
    for cfg in activeConfigs(): os.remove(cfg)


################################################################################
def loadHostConfig(timeout=15):
    start = time.time()
    while time.time() - start < timeout:
        for cfgName in activeConfigs():
            if re.search("^host", cfgName.split(os.sep)[-1]):
                cfg = Config()
                cfg.load(cfgName)
                return cfg
    raise c.TimeoutExceeded("could not identify the host config within %d seconds."%(timeout))


################################################################################
def getSlaveConfigs(numConfigs=1, timeout=15):
    ret = set()
    start = time.time()
    while len(ret) < numConfigs: # keep looking for slave configs until they're found or timeout
        if time.time() - start > timeout:
            raise c.TimeoutExceeded("could not identify any slave configs"\
                "within %d seconds"%(timeout))
        for cfgName in activeConfigs():
            if re.search("^host", cfgName): continue
            cfg = Config(host=True)
            cfg.load(cfgName)
            ret.add( cfg )
    return ret


################################################################################
class Config(object):
    """the grand collection of information that determines how a Starcraft 2
    game is setup and intended to behave"""
    ############################################################################
    def __init__(self, load=None,
        expo=c.DEFAULT_EXPANSION, version=None,
        ladder=None, players=[], whichPlayer="", mode=None, themap=None, numObservers=0, start=None, trust=True,
        ipAddress=[], ports=[], host=[], slaves=[],
        fogDisabled=False, stepSize=0, opponents=[], fullscreen=True,
        raw=False, score=False, feature=False, render=False,
        replay=None, debug=False,
            #connects=[],
            #state=c.GAME_INIT,
            #agentRaces=[], bots=[],
    ):
        self._gotPorts  = False # ensure that ports are only returned if they were retrieved on this machine
        # flexible settings auto-determination
        if isinstance(players, list):       pass
        elif hasattr(players, "__iter__"):  players = list(players)
        else:                               players = [players] # wrap
        # application setup info
        self.expo       = expo
        self.version    = version # automatically calculate unless specified
        # match setup info
        self.ladder     = ladder
        self.players    = players
        self.thePlayer  = whichPlayer
        self.mode       = mode
        self.themap     = themap
        self.numObserve = int(numObservers)
        self.start      = start  # indicates the final match configuration has been determined
        self.trustOpps  = trust  # request secure game hosting on an official ladder machine
        # connection info
        self.ipAddress  = ipAddress
        self.ports      = ports
        self.host       = host   # if defined, this config is for a slave
        self.slavePorts = slaves # if not defined, this is for the host
        self.ladderMsg  = ""
        # pre-game behavior
        self.scenario   = None
        # in-game behavior
        self.fogDisabled= fogDisabled
        self.stepSize   = int(stepSize)
        self.opponents  = opponents # names of specific opponents
        self.fullscreen = fullscreen
        # observation data content
        self.raw        = raw
        self.score      = score
        self.feature    = feature
        self.render     = render
        # post-game behavior
        self.replay     = replay
        self.reqFiles   = []
        # class behavior
        self.debug      = debug
        #self.state = c.GAME_INIT
        if load:    self.load(load) # specific configuration filename to load
        else:       self.inflate()  # ensure all objects are objects
        self.isCustom   = bool(self.opponents or self.fogDisabled or \
                               self.themap    or self.stepSize)
    ############################################################################
    def __str__(self):  return self.__repr__()
    def __repr__(self):
        name = self.__class__.__name__
        if self.slavePorts: ip = [str(x) for x in self.host] # this configuration is a host configuration when slave ports are defined
        else:               ip = [str(x) for x in self.connection]
        ipStr = " @ %s"%(":".join(ip)) if ip else ""
        return "<%s %s %s%s>"%(name, self.installedApp.name(), self.version, ipStr)
    ############################################################################
    def __del__(self):
        self.returnPorts()
        try:    self.disable()
        except IOError: pass
    ############################################################################
    @property
    def attrs(self):
        data = dict(self.__dict__) # a new, altered dict is returned
        for k,v in sorted(iteritems(data)):
            #if re.search(Config.regexRaw, k): continue # ignore attrs that aren't raw data
            if re.search("^_", k):  del data[k]
        return data
    ############################################################################
    @property
    def agents(self):
        """identify which players are agents (not observers or computers). Errors if flattened."""
        ret = []
        for player in self.players:
            if player.isComputer: continue
            try:
                if player.observer: continue
            except: pass
            ret.append(player)
        return ret
    ############################################################################
    @property
    def agentRaces(self):
        """identify the races of the players that are agents. Errors if flattened."""
        return [player.selectedRace for player in self.players if not player.isComputer]
    ############################################################################
    @property
    def computers(self):
        """identify the players that are Blizzard's built-in bots. Errors if flattened."""
        return [player for player in self.players if player.isComputer]
    ############################################################################
    @property
    def allLobbySlots(self):
        """the current configuration of the lobby's players, defined before the match starts"""
        if self.debug:
            p      = ["Lobby Configuration detail:"] + \
                     ["    %s:%s%s"%(p, " "*(12-len(p.type)), p.name)]
                     #["    agent:     %s"%p for p in self.agents] + \
                     #["    computer:  %s, %s"%(r,d) for r,d in self.computers]
            if self.observers: # must separate condition because numObs is a number, not an iterator
                p += ["    observers: %d"%self.observers]
            print(os.linesep.join(p))
        return (self.agents, self.computers, self.observers)
    ############################################################################
    @property
    def connection(self):
        """identify the remote connection parameters"""
        self.getPorts()         # acquire if necessary
        self.getIPaddresses()   # acquire if necessary
        return (self.ipAddress, self.ports)
    ############################################################################
    @property
    def execPath(self):
        """the executable application's path"""
        vers = self.version.label if self.version else None # executables in Versions folder are stored by baseVersion (modified by game data patches)
        return self.installedApp.exec_path(vers)
    ############################################################################
    @property
    def is64bit(self):
        """whether the this machine is 64-bit capable or not"""
        return self.installedApp.is64bit
    ############################################################################
    @property
    def installedApp(self):
        """identify the propery application to launch, given the configuration"""
        try:    return self._installedApp
        except: # raises if not yet defined
            self._installedApp = runConfigs.get() # application/install/platform management
            return self._installedApp
    ############################################################################
    @property
    def interfaces(self):
        return (self.raw    , self.score,
                self.feature, self.render)
    ############################################################################
    @property
    def isHost(self):
        """is the host if it lacks host details"""
        return not bool(self.host) and bool(self.slavePorts)
    ############################################################################
    @property
    def isMultiplayer(self):
        return self.numGameClients > 1
    ############################################################################
    @property
    def mapData(self):
        """the raw, binary contents of the Starcraft2 map file"""
        return self.themap.rawData
        #with open(self.themap.path, "rb") as f:
        #    return f.read()
    ############################################################################
    @property
    def mapLocalPath(self):
        return self.themap.path
    ############################################################################
    @property
    def name(self):
        #try:    return self.cfgFile
        #except: pass # else extract name from loaded cfg name
        #k = "host" if self.host else os.getpid()
        #self.cfgFile = os.path.join(c.FOLDER_ACTIVE_CONFIGS, "%s_%s.json"%(k, dateFormat.now()))
        #return self.cfgFile
        return os.path.join(
                    c.FOLDER_ACTIVE_CONFIGS,
                    "%s_%s.json"%(os.getpid(), dateFormat.now(self.start)))
    ############################################################################
    @property
    def numAgents(self):
        return len(self.agents)
    ############################################################################
    @property
    def numBots(self):
        return len(self.computers)
    ############################################################################
    @property
    def numGameClients(self):
        """the number of game client connections in the match"""
        return self.numAgents + self.numObserve
    ############################################################################
    @property
    def numPlayers(self):
        #return self.numAgents + self.numBots
        return len(self.players)
    ############################################################################
    @property
    def observers(self):
        """the players who are actually observers"""
        ret = []
        for player in self.players:
            try:
                if player.observer: ret.append(player)
            except: pass # ignore PlayerRecords which don't have an observer attribute
        return ret
    ############################################################################
    @property
    def os(self):
        """the operating system this match is loaded on"""
        return self.installedApp.name()
    ############################################################################
    @property
    def participants(self):
        """agents + computers (i.e. all non-observers)"""
        ret = []
        for p in self.players:
            try:
                if     p.isComputer: ret.append(p)
                if not p.isObserver: ret.append(p) # could cause an exception if player isn't a PlayerPreGame
            except AttributeError: pass
        return ret
    ############################################################################
    @property
    def realtime(self):
        """whether the game progresses in realtime or requires each player to
        summit step requests to advance the game"""
        return not bool(self.stepSize)
    ############################################################################
    @property
    def teams(self):
        players = self.participants
        div     = len(players) / 2.
        teamSize= int(div)
        if (div - teamSize) > 0.49: # if odd number of participants
            teamSize += 1 # if remainder, round up
        return (players[:teamSize], players[teamSize:])
    ############################################################################
    def addPlayer(self, player):
        if not isinstance(player, PlayerPreGame):
            raise ValueError("%s must be a %s"%(player, PlayerPreGame))
        self.players.append(player)
    ############################################################################
    def clientInitHost(self):
        """extract ipAddress information used by the local client to connect to another SC2 client at this IP address"""
        if self.host:   return self.host[0]
        else:           return self.getIPaddresses()[-1]
    ############################################################################
    def clientInitPort(self):
        """extract the port that is used to connect to the locally launched SC2 client"""
        ports = self.getPorts()
        return ports[2] # private shared_port / init port
    ############################################################################
    def disable(self):
        try:
            os.remove(self.name) # delete the file
            #del self.cfgFile        # forget the cfg file's name
        except: pass
    ############################################################################
    def display(self, header=""):
        print(self)
        h = "[%s]  "%header.upper() if header else ""
        for k,v in sorted(iteritems(self.attrs)):
            print("%s%16s : %s"%(h,k,v))
    ############################################################################
    def flatten(self, data=None):
        """reduce all objects into simplified values as a attr dictionary that
        could be transformed back into a full configuration via inflate()"""
        if data == None: data=self.attrs
        ret = {}
        for k,v in iteritems(data):
            if not v: continue # don't flatten if there's nothing to flatten
            elif k == "expo":               v = v.type
            elif k == "version":            v = v.label
            elif k == "ladder":             v = v.name
            elif k == "players":
                newPs = []
                for i,p in enumerate(v):
                    try:    diff = p.difficulty.type
                    except: diff = p.difficulty
                    if isinstance(p, PlayerPreGame):    newPs.append( (p.name, p.type.type, p.initCmd, p.initOptions, diff, p.rating, p.selectedRace.type, self.numObserve, p.playerID, p.raceDefault) )
                    else:                               newPs.append( (p.name, p.type.type, p.initCmd, p.initOptions, diff, p.rating) )
                # TODO -- handle if type or observers params are not available (i.e. if a simple PlayerRecord, not a PlayerPreGame
                ret[k] = newPs
                continue
            elif k == "mode"   and self.mode:   v = v.type
            #elif k == "state":
            elif k == "themap" and self.themap: v = v.name
            ret[k] = v
        return ret
    ############################################################################
    def inflate(self, newData={}):
        """ensure all object attribute values are objects"""
        self.__dict__.update(newData)
        #if not isinstance(self.state, types.GameStates):      self.state     = types.GameStates(self.state)
        if self.expo    and not isinstance(self.expo, types.ExpansionNames):    self.expo       = types.ExpansionNames(self.expo)
        if self.version and not isinstance(self.version, versions.Version):     self.version    = versions.Version(self.version)
        if self.ladder  and not isinstance(self.ladder, Ladder):                self.ladder     = Ladder(self.ladder)
        for i,player in enumerate(self.players): # iterate over all players
            if       isinstance(player, str):                                   self.players[i] = getPlayer(player)
            elif not isinstance(player, PlayerRecord):                          self.players[i] = buildPlayer(*player)
        if self.mode    and not isinstance(self.mode, types.GameModes):         self.mode       = types.GameModes(self.mode)
        if self.themap  and not isinstance(self.themap, MapRecord):             self.themap     = selectMap(name=self.themap)
    ############################################################################
    def launchApp(self, **kwargs):
        """Launch Starcraft2 process in the background using this configuration.
        WARNING: if the same IP address and port are specified between multiple
                 SC2 process instances, all subsequent processes after the first
                 will fail to initialize and crash.
        """
        app = self.installedApp
        # TODO -- launch host in window minimized/headless mode
        vers = self.getVersion()
        return app.start(version=vers,#game_version=vers.baseVersion, data_version=vers.dataHash,
            full_screen=self.fullscreen, verbose=self.debug, **kwargs)
    ############################################################################
    def load(self, cfgFile=None, timeout=None):
        """expect that the data file has already been established"""
        #if cfgFile != None: self.cfgFile = cfgFile # if it's specified, use it
        if not cfgFile:
            cfgs = activeConfigs()
            if   len(cfgs) > 1: raise Exception("found too many configurations (%s); not clear which to load: %s"%(len(cfgs), cfgs))
            elif len(cfgs) < 1:
                if timeout: # wait for a configuration file to appear to be loaded
                    startWait = time.time()
                    timeReported = 0
                    while not cfgs:
                        timeWaited = time.time() - startWait
                        if timeWaited > timeout:
                            raise c.TimeoutExceeded("could not join game after %s seconds"%(timeout))
                        try:  cfgs = activeConfigs()
                        except:
                            if self.debug and timeWaited - timeReported >= 1:
                                timeReported += 1
                                print("second(s) waited for game to appear:  %d"%(timeReported))
                else:  raise Exception("must have a saved configuration to load or allow loading via timeout setting")
            cfgFile = cfgs.pop()
        try:
            with open(cfgFile, "rb") as f:
                data = f.read() # bytes => str
        except TypeError as e:
            print("ERROR %s: %s %s"%(e, cfgFile, type(cfgFile)))
            raise
        self.loadJson(data) # str => dict
        if self.debug:
            print("configuration loaded: %s"%(self.name))
            self.display()
    ############################################################################
    def loadJson(self, data):
        """convert the json data into updating this obj's attrs"""
        if not isinstance(data, dict):
            data = json.loads(data)
        self.__dict__.update(data)
        self.inflate() # restore objects from str values
        #if self.ports:  self._gotPorts = True
        return self
    ############################################################################
    def toJson(self, data=None, pretty=False):
        """convert the flattened dictionary into json"""
        if data==None: data = self.attrs
        data = self.flatten(data) # don't send objects as str in json
        #if pretty:
        ret = json.dumps(data, indent=4, sort_keys=True)
        #self.inflate() # restore objects from json str data
        return ret
    ############################################################################
    def getVersion(self):
        """the executable application's version"""
        if isinstance(self.version, versions.Version):  return self.version
        if self.version: # verify specified version exists
            version = versions.Version(self.version) # create this object to allow self._version_ to be specified in multiple different ways by the user
            if version.baseVersion not in self.installedApp.versionMap(): # verify that the selected version has an executable
                raise runConfigs.lib.SC2LaunchError(
                    "specified game version %s executable is not available.%s    available:  %s"%( \
                    version, os.linesep, "  ".join(self.installedApp.listVersions())))
            self.version = version
        else: # get most recent executable's version
            path = self.installedApp.exec_path()
            vResult = self.installedApp.mostRecentVersion
            self.version = versions.Version(vResult)
        if self.debug: print(os.linesep.join([
            "Game configuration detail:",
            "    platform:   %s"%(self.os),
            "    app:        %s"%(self.execPath),
            "    version:    %s"%(self.version)]))
        return self.version
    ############################################################################
    def getIPaddresses(self):
        """identify the IP addresses where this process client will launch the SC2 client"""
        if not self.ipAddress:
            self.ipAddress = ipAddresses.getAll() # update with IP address
        return self.ipAddress
    ############################################################################
    def getPorts(self):
        """acquire ports to be used by the SC2 client launched by this process"""
        if self.ports: # no need to get ports if ports are al
            return self.ports
        if not self._gotPorts:
            self.ports = [
                portpicker.pick_unused_port(), # game_port
                portpicker.pick_unused_port(), # base_port
                portpicker.pick_unused_port(), # shared_port / init port
            ]
            self._gotPorts = True
        return self.ports
    ############################################################################
    def requestCreateDetails(self):
        """add configuration to the SC2 protocol create request"""
        createReq = sc_pb.RequestCreateGame( # used to advance to Status.initGame state, when hosting
            realtime    = self.realtime,
            disable_fog = self.fogDisabled,
            random_seed = int(time.time()), # a game is created using the current second timestamp as the seed
            local_map   = sc_pb.LocalMap(map_path=self.mapLocalPath,
                                         map_data=self.mapData))
        for player in self.players:
            reqPlayer = createReq.player_setup.add() # add new player; get link to settings
            playerObj = PlayerPreGame(player)
            if playerObj.isComputer:
                reqPlayer.difficulty    = playerObj.difficulty.gameValue()
            reqPlayer.type              = c.types.PlayerControls(playerObj.control).gameValue()
            reqPlayer.race              = playerObj.selectedRace.gameValue()
        return createReq # SC2APIProtocol.RequestCreateGame
    ############################################################################
    def requestJoinDetails(self):
        """add configuration information to the SC2 protocol join request
    REQUIREMENTS FOR SUCCESSFUL LAUNCH:
    server game_port must match between all join requests to client (represents the host's port to sync game clients)
    server base_port
    client game_port must be unique between each client (represents ???)
    client base_port
    client shared_port must match between all join requests to client
        """
        raw,score,feature,rendered = self.interfaces
        interface = sc_pb.InterfaceOptions()
        interface.raw   = raw   # whether raw data is reported in observations
        interface.score = score # whether score data is reported in observations
        interface.feature_layer.width = 24
        #interface.feature_layer.resolution =
        #interface.feature_layer.minimap_resolution =
        joinReq = sc_pb.RequestJoinGame(
            options = interface,
            #observed_player_id=__?__,
            race = self.whoAmI().selectedRace.gameValue())
        # TODO -- allow player to be an observer, not just a player w/ race
        if self.host: # always add ports for joining player to connect to defined host
            hostPorts = self.host[1]
            joinReq.server_ports.game_port = hostPorts[0]
            joinReq.server_ports.base_port = hostPorts[1]
            joinReq.shared_port            = hostPorts[2]
            clientPorts = joinReq.client_ports.add()
            clientPorts.game_port = self.ports[0]
            clientPorts.base_port = self.ports[1]
            ret = self.ports[2]
        elif self.isMultiplayer: # always add ports as host of multiple agents/clients
            if len(self.ports) < 5:
                self.ports += [ # get new private client ports for the host
                    portpicker.pick_unused_port(), # game_port
                    portpicker.pick_unused_port(), # base_port
                ]
            joinReq.server_ports.game_port = self.ports[0]
            joinReq.server_ports.base_port = self.ports[1]
            joinReq.shared_port            = self.ports[2]
            clientPorts = joinReq.client_ports.add()
            clientPorts.game_port = self.ports[3] # new private client game port
            clientPorts.base_port = self.ports[4] # new private client base port
        return joinReq # SC2APIProtocol.RequestJoinGame
    ############################################################################
    def returnPorts(self):
        """deallocate specific ports on the current machine"""
        if self._gotPorts:
            #print("deleting ports >%s<"%(self.ports))
            map(portpicker.return_port, self.ports)
            self._gotPorts = False
        self.ports = []
    ############################################################################
    def save(self, filename=None, debug=False):
        """save a data file such that all processes know the game that is running"""
        if not filename: filename = self.name
        with open(filename, "w") as f: # save config data file
            f.write(self.toJson(self.attrs))
        if self.debug or debug:
            print("saved configuration %s"%(self.name))
            for k,v in sorted(iteritems(self.attrs)):
                print("%15s : %s"%(k,v))
    ############################################################################
    def updateIDs(self, ginfo, tag=None, debug=False):
        """ensure all player's playerIDs are correct given game's info"""
            # SC2APIProtocol.ResponseGameInfo attributes:
                # map_name
                # mod_names
                # local_map_path
                # player_info
                # start_raw
                # options
        thisPlayer = self.whoAmI()
        for pInfo in ginfo.player_info: # parse ResponseGameInfo.player_info to validate player information (SC2APIProtocol.PlayerInfo) against the specified configuration
            pID = pInfo.player_id
            if pID == thisPlayer.playerID: continue # already updated
            pCon = c.types.PlayerControls(pInfo.type)
            rReq = c.types.SelectRaces(pInfo.race_requested)
            for p in self.players: # ensure joined player is identified appropriately
                if p.playerID and p.playerID != pID: continue # if this non-matching player already has a set playerID, it can't match
                if p.control == pCon and p.selectedRace == rReq: # matched player
                    p.playerID = pID # updated player IDs should be saved into the game configuration
                    if debug: print("[%s] match contains %s."%(tag, p))
                    pID = 0 # declare that the player has been identified
                    break
            if pID: raise c.UnknownPlayer("could not match %s %s %s to any "
                "existing player of %s"%(pID, pCon, rReq, self.players))
    ############################################################################
    def whoAmI(self):
        """return the player object that owns this configuration"""
        self.inflate() # ensure self.players contains player objects
        if self.thePlayer:
            for p in self.players:
                if p.name != self.thePlayer: continue
                return p
        elif len(self.players) == 1:
            ret = self.players[0]
            self.thePlayer = ret.name # remember this for the future in case more players are added
            return ret
        raise Exception("could not identify which player this is given %s (%s)"%(self.players, self.thePlayer))





import multiprocessing # ensure that multiProcessing.Queue can be instantiated in a child process


################################################################################
def doNothing(observation):
    """simply ignore the observation
    default callaback unless specified otherwise.
    useful if a human is playing unaided
    """


################################################################################
class forwardObservation(object):
    """allow the observation getter to connect to the commander by forwarding observations"""
    ############################################################################
    def __init__(self, pushQueue):
        self.pushQ = pushQueue
    ############################################################################
    def __call__(self, observation):
        self.pushQ.put(observation)





"""
PURPOSE: front-end interface to easily and reliably match against opponents and
         run Starcraft2 opponents.
"""

import argparse
import sys

from sc2common import types as t
from sc2ladderMgmt import getKnownLadders
from sc2maptool.functions import filterMapNames
from sc2players import getKnownPlayers

from sc2gameLobby.__version__ import __version__
from sc2gameLobby import gameConstants as c
from sc2gameLobby import launcher


################################################################################
def optionsParser(passedParser=None):
    if passedParser == None:
        parser = argparse.ArgumentParser(
           #usage="python %s"%__file__,
            description=__doc__,
            epilog="version: %s"%__version__)
    else:
        parser = passedParser
    ALLOWED_PLAYERS = list(getKnownPlayers())
    ALLOWED_LADDERS = list(getKnownLadders())
    ALLOWED_MAPS    = filterMapNames("", closestMatch=False)
    gameOptions = parser.add_argument_group('game lobby actions')
    gameOptions.add_argument("--nogui"          , action="store_true"   , help="launch game directly using command-line arguments formatted as key=value.")
    gameOptions.add_argument("--search"         , default=""            , help="retrieve player information from the ladder (comma separated names)", metavar="PLAYERS")
    gameOptions.add_argument("--history"        , action="store_true"   , help="include match history when using the --search option.")
    startOption = parser.add_argument_group('sc2 game client gameplay options')
    startOption.add_argument("-l", "--ladder"   , default="versentiedge", choices=ALLOWED_LADDERS
                                                                        , help="the ladder where the game is run.  Allowed values: " + ", ".join(ALLOWED_LADDERS), metavar='')
    startOption.add_argument("-p", "--player"   , default="defaulthuman", choices=ALLOWED_PLAYERS
                                                                        , help="the player profile the ladder identifies you with.  Allowed values: " + ", ".join(ALLOWED_PLAYERS), metavar='')
    startOption.add_argument("-v", "--version"  , type=int              , help="the specific game version to play.", metavar='')
    startOption.add_argument("-e", "--exp"      , default="lotv", choices=c.EXPO_SELECT.keys(), metavar=''
                                                                        , help="the specific game expansion version to play.  Allowed values: " + ", ".join(c.EXPO_SELECT.keys()))
    startOption.add_argument("-r", "--race"     , default=c.RANDOM, choices=t.SelectRaces.ALLOWED_TYPES, metavar=''
                                                                        , help="selected race to play. Allowed values: " + ", ".join(t.SelectRaces.ALLOWED_TYPES))
    startOption.add_argument("-m", "--mode"     , default=c.MODE_1V1, choices=t.GameModes.ALLOWED_TYPES, metavar=''
                                                                        , help="game mode to play. DEFAULT: %s"%c.MODE_1V1)
    startOption.add_argument("-o", "--obs"      , default=0, type=int   , help="the number of observers that will watch this match for this player", metavar='INT')
    localOption = parser.add_argument_group("this individual sc2 client's options")
    localOption.add_argument("--observe"        , action="store_true"   , help="be a match observer, not a participant.")
   #localOption.add_argument("--replay"         , default=""            , help="reference a replay file as input.", metavar="FILE")
    localOption.add_argument("--savereplay"     , default=""            , help="store a replay as the name (and/or path) of the output replay file after the match finishes.", metavar="NAME")
   #localOption.add_argument("--simulate"       , default=""            , help="simulation setup files for the players to play", metavar="FILE_LIST")
    localOption.add_argument("--windowed"       , action="store_true"   , help="launch client in windowed mode (default: full screen).")
    noLadderOpt = parser.add_argument_group('mandate a non-ladder (custom) match')
    noLadderOpt.add_argument("--opponents"      , default=""            , help="specify specific opponent(s), comma separated.", metavar='LIST')
    noLadderOpt.add_argument("--nofog"          , action="store_true"   , help="disable fog of war.")
    noLadderOpt.add_argument("--map"            , default=None          , help="play on a specific map.", metavar='MAPNAME')
    noLadderOpt.add_argument("--step"           , default=0, type=int   , help="the number of steps between step() actions.  A zero value means that the game runs in realtime (default)", metavar='INT')
    aiBotOption = parser.add_argument_group("observation response contents for AI / bots")
    aiBotOption.add_argument("--raw"            , action="store_true"   , help="return raw data in observations")
    aiBotOption.add_argument("--feature"        , action="store_true"   , help="return feature data in observations.")
    aiBotOption.add_argument("--rendered"       , action="store_true"   , help="return fully rendered data in observations.")
    aiBotOption.add_argument("--score"          , action="store_true"   , help="return current score data in observations.")
    aiBotOption.add_argument("--resolution"     , default=""            , help="resolution that the feature/rendered data is returned in.", metavar='AxB')
    aiBotOption.add_argument("--minimap"        , default=""            , help="resolution that the feature/rendered minimap data is returned in.", metavar='AxB')
    aiBotOption.add_argument("--layerwidth"     , type=int, default=24  , help="the mapping of number of pixels to game grid coordinates.", metavar='INT')
    versionOpts = parser.add_argument_group('sc2 game client version handling')
    versionOpts.add_argument("--update"         , type=str,default=""   , help="update an existing version.", metavar='<label>,<version>,<base-version>')
    versionOpts.add_argument("--add"            , type=str,default=""   , help="add a new version.", metavar='<label>,<version>,<base-version>')
    versionOpts.add_argument("--versions"       , action="store_true"   , help="display known Starcraft 2 public versions.")
    return parser


################################################################################
def main(options=None):
    if options == None:
        options = optionsParser().parse_args()
        options.scenario = None # hardcode this value because no scenario is created in regular gameplay, nor is such an option provided
        sys.argv = sys.argv[:1] # remove all arguments to avoid problems with absl FLAGS :(
    launcher.run(options)




#
# Copyright 2018 Versentiedge Inc. All Rights Reserved.
#   *** partially modified version previously published by Google Inc. ***
#   - revised logic when selecting the executable version
#   - now relies on external game version management module
#   - added mapsDir to ensure game knows where to find maps
#   - remove start() from LocalBase to allow a pure passthrough to lib.RunConfig
#   - added versionsDir to consolidate version dir lookups
#   - added validVersionExecutables to get all game versions available on this machine
#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configs for how to run SC2 from a normal install on various platforms."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six import iteritems # python 2/3 compatibility

import copy
import os
import platform
import sys

from sc2gameLobby import versions
from pysc2.run_configs import lib
from pysc2.lib import sc_process


################################################################################
class LocalBase(lib.RunConfig):
  """Base run config for the deepmind file hierarchy."""
  ##############################################################################
  def __init__(self, base_dir, exec_name, cwd=None, env=None):
    base_dir = os.path.expanduser(base_dir)
    cwd = cwd and os.path.join(base_dir, cwd)
    super(LocalBase, self).__init__(
        replay_dir=os.path.join(base_dir, "Replays"),
        data_dir=base_dir, tmp_dir=None, cwd=cwd, env=env)
    self._exec_name = exec_name
    self.xyz = None
  ##############################################################################
  @property
  def is64bit(self):
      """whether the this machine is 64-bit capable or not"""
      return platform.machine().endswith('64')
  ##############################################################################
  @property
  def mapsDir(self):
      return os.path.join(self.data_dir, "Maps")
  ##############################################################################
  @property
  def mostRecentVersion(self):
      versMap = self.versionMap()
      orderedVersions = sorted(list(iteritems(versMap)))
      #vers = [v[-1] for k,v in sorted(iteritems(self.versionMap()))]
      return orderedVersions[-1][-1]
  ##############################################################################
  @property
  def validVersionExecutables(self):
      ret = []
      for x in os.listdir(self.versionsDir):
          if x.startswith("Base"):
              ret.append(int(x[4:]))
      return ret
  ##############################################################################
  @property
  def versionsDir(self):
      return os.path.join(self.data_dir, "Versions")
  ##############################################################################
  def exec_path(self, baseVersion=None):
    """Get the exec_path for this platform. Possibly find the latest build."""
    if not os.path.isdir(self.data_dir):
        raise sc_process.SC2LaunchError("Install Starcraft II at %s or set the SC2PATH environment variable"%(self.data_dir))
    if baseVersion==None: # then select most recent version's baseVersion
        mostRecent = versions.handle.mostRecent
        if mostRecent:  return mostRecent["base-version"]
        raise sc_process.SC2LaunchError(
            "When requesting a versioned executable path without specifying base-version, expected "
            "to find StarCraft II versions installed at %s."%(self.versionsDir))
    elif isinstance(baseVersion, versions.Version):
        baseVersion = baseVersion.baseVersion
    elif str(baseVersion).count(".") > 0:
        baseVersion = versions.Version(baseVersion).baseVersion
    #else: # otherwise expect that the baseVersion specified is correct
    baseVersExec = os.path.join(self.versionsDir, "Base%s"%baseVersion, self._exec_name)
    if os.path.isfile(baseVersExec):
        return baseVersExec # if baseVersion in Versions subdir is valid, it is the correct executable
    raise sc_process.SC2LaunchError("Specified baseVersion %s does not exist at %s.%s    available: %s"%(\
        baseVersion, baseVersExec, os.linesep, " ".join(
            str(val) for val in sorted(self.versionMap().keys())) ))
  ##############################################################################
  def listVersions(self):
      ret = []
      map(ret.extend, self.versionMap().values())
      return sorted(ret)
  ##############################################################################
  def versionMap(self, debug=False):
      ret = {}
      for vKey in self.validVersionExecutables:
          labels = [r["label"] for r in versions.handle.search(**{"version":vKey})]
          ret[vKey] = labels
      return ret
      #return [versions.Version(vKey).label for vKey in self.validVersionExecutables]
  ##############################################################################
  def start(self, version=None, **kwargs):#game_version=None, data_version=None, **kwargs):
    """Launch the game process."""
    if not version:
        version = self.mostRecentVersion
    pysc2Version = lib.Version( # convert to pysc2 Version
        version.version,
        version.baseVersion,
        version.dataHash,
        version.fixedHash)
    return sc_process.StarcraftProcess(
                self,
                exec_path=self.exec_path(version.baseVersion),
                version=pysc2Version,
                **kwargs)


################################################################################
class Windows(LocalBase):
  """Run on Windows."""
  ##############################################################################
  def __init__(self):
    super(Windows, self).__init__(
        os.environ.get("SC2PATH", "C:/Program Files (x86)/StarCraft II").strip(
        '"'), "SC2_x64.exe", "Support64")
  ##############################################################################
  @classmethod
  def priority(cls):
    if platform.system() == "Windows":
      return 1


################################################################################
class MacOS(LocalBase):
  """Run on MacOS."""
  ##############################################################################
  def __init__(self):
    super(MacOS, self).__init__(
        os.environ.get("SC2PATH", "/Applications/StarCraft II"),
        "SC2.app/Contents/MacOS/SC2")
  ##############################################################################
  @classmethod
  def priority(cls):
    if platform.system() == "Darwin":
      return 1


################################################################################
class Linux(LocalBase):
  """Config to run on Linux."""
  ##############################################################################
  def __init__(self):
    base_dir = os.environ.get("SC2PATH", "~/StarCraftII")
    base_dir = os.path.expanduser(base_dir)
    env = copy.deepcopy(os.environ)
    env["LD_LIBRARY_PATH"] = ":".join(filter(None, [
        os.environ.get("LD_LIBRARY_PATH"),
        os.path.join(base_dir, "Libs/")]))
    super(Linux, self).__init__(base_dir, "SC2_x64", env=env)
  ##############################################################################
  @classmethod
  def priority(cls):
    if platform.system() == "Linux":
      return 1





from sc2gameLobby import gameConstants as c


################################################################################
def tieDetermined(cfg):
    """all players tied their results"""
    return assignValue(cfg, c.RESULT_TIE, c.RESULT_TIE)


################################################################################
def launchFailure(cfg):
    """report that the match failed to start as anticipated"""
    return assignValue(cfg, c.RESULT_UNDECIDED, c.RESULT_UNDECIDED)


################################################################################
def playerSurrendered(cfg):
    """the player has forceibly left the game"""
    if cfg.numAgents + cfg.numBots == 2:
          otherResult = c.RESULT_VICTORY
    else: otherResult = c.RESULT_UNDECIDED # if multiple players remain, they need to finish the match
    return assignValue(cfg, c.RESULT_DEFEAT, otherResult)


################################################################################
def playerCrashed(cfg):
    """occurs when a player's understood state hasn't changed for too long"""
    return assignValue(cfg, c.RESULT_CRASH, c.RESULT_UNDECIDED)


################################################################################
def playerDisconnected(cfg):
    """occurs when a game client loses its connection from the host"""
    return assignValue(cfg, c.RESULT_DISCONNECT, c.RESULT_UNDECIDED)


################################################################################
def assignValue(cfg, playerValue, otherValue):
    """artificially determine match results given match circumstances.
    WARNING: cheating will be detected and your player will be banned from server"""
    player = cfg.whoAmI()
    result = {}
    for p in cfg.players:
        if p.name == player.name:   val = playerValue
        else:                       val = otherValue
        result[p.name] = val
    return result 


################################################################################
def idPlayerResults(cfg, rawResult):
    """interpret standard rawResult for all players with known IDs"""
    result = {}
    knownPlayers = []
    dictResult = {plyrRes.player_id : plyrRes.result for plyrRes in rawResult}
    for p in cfg.players:
        if p.playerID and p.playerID in dictResult: # identified player w/ result
            knownPlayers.append(p)
            result[p.name] = dictResult[p.playerID]
    #if len(knownPlayers) == len(dictResult) - 1: # identified all but one player
    #    for p in cfg.players: # search for the not identified player
    #        if p in knownPlayers: continue # already found
    #        result.append( [p.name, p.playerID, dictResult[p.playerID]] )
    #        break # found missing player; stop searching
    #for r in result:
    #    print("result:>", r)
    return result





"""PURPOSE: launch the mini-editor to create custom p setups"""

import glob
import os
import stat
import subprocess

from sc2gameLobby import debugCmds
from sc2gameLobby import gameConstants as c
from sc2gameLobby.gameConfig import Config

# windows 10
banksDir = "C:\\Users\\jared\\Documents\\StarCraft II\\Banks"
availableBankNames = glob.glob(os.path.join(banksDir, "*"))
editorCmd = "%s -run %s -testMod %s -displayMode 1"
#EXAMPLE:
#C:\\StarCraft II\\Support64>SC2Switcher_x64.exe -run "Ladder\\CatalystLE.SC2Map" -testMod "C:\\Users\\TheWisp\\Documents\\PlayGround.SC2Mod"

################################################################################
def launchEditor(mapObj):
    """
    PURPOSE: launch the editor using a specific map object
    INPUT:   mapObj (sc2maptool.mapRecord.MapRecord)
    """
    cfg = Config()
    if cfg.is64bit: selectedArchitecture = c.SUPPORT_64_BIT_TERMS
    else:           selectedArchitecture = c.SUPPORT_32_BIT_TERMS
    fullAppPath = os.path.join(
        cfg.installedApp.data_dir,
        c.FOLDER_APP_SUPPORT%(selectedArchitecture[0]))
    appCmd = c.FILE_EDITOR_APP%(selectedArchitecture[1])
    fullAppFile = os.path.join(fullAppPath, appCmd)
    os.chmod(fullAppFile, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IXUSR|\
                          stat.S_IRUSR|stat.S_IWUSR|stat.S_IWGRP|stat.S_IXGRP) # switcher file must be executable
    finalCmd = editorCmd%(appCmd, mapObj.path, c.FILE_EDITOR_MOD)
    cwd = os.getcwd()
    os.chdir(fullAppPath)
    os.system(finalCmd)
    os.chdir(cwd) # restore back to original directory


################################################################################
def initScenario(controller, scenario):
    """once in the in_game state, use the controller to set up the scenario"""
    knownUnits = []
    gl = 0
    getGameState = controller.observe # function that observes what's changed since the prior gameloop(s)
    scenarioUnits = list(scenario.units.values())
    while True:
        obs = getGameState()
        if obs.observation.game_loop <= 1: continue
        knownUnits = obs.observation.raw_data.units # identify existing units (requieres no fog to be disabled?)
        if knownUnits: break # wait until units are found in the raw observation
    createCmds = debugCmds.create(*scenarioUnits)
    controller.debug(*createCmds) # send create cmd via controller
    rmTags = []
    keepUnits = {18, 59, 86} # command center, nexus and hatchery
    for unit in list(knownUnits):
        if unit.alliance == c.NEUTRAL:  continue # ignore neutral units visible via snapshot to start the game
        if unit.mineral_contents:       continue # don't remove mineral nodes
        if unit.vespene_contents:       continue # don't remove vespene nodes
        if unit.unit_type in keepUnits: continue # don't remove the main building
        rmTags.append(unit.tag)
    if rmTags:
        rmCmd = debugCmds.remove(*rmTags) # create command to remove existing units
        controller.debug(rmCmd) # send remove cmd to remove existing units
    newUnits = {}
    triesRemaining = 15
    while len(newUnits) < len(scenarioUnits) and triesRemaining > 0: # wait until new units are created
        units = getGameState().observation.raw_data.units
        for i, unit in enumerate(scenarioUnits): # match new unit tags with their originating units
            if unit.tag: continue # already found a matching tag for this unit
            for liveUnit in units: # identify new units and their tags
                if liveUnit.unit_type != unit.code:  continue # can't match a unit of a different type
                if liveUnit.owner     != unit.owner: continue # can't match units with different owners
                if liveUnit.tag       in newUnits:   continue # don't match the same unit twice
                unit.tag = liveUnit.tag # found a match; sync tags + decalre this unit matched
                newUnits[unit.tag] = unit # remember this association between unit and its properties
                break # once a match is made, stop searching
            if not unit.tag:
                print("%02d.  missing: %s"%(15-triesRemaining, unit))
        triesRemaining -= 1
    for term in newUnits:
        print(term)
    modifyCmds = debugCmds.modify(*newUnits.values()) # create command to set properties of in-game units
    #controller.debug(*modifyCmds) # send modify cmd via controller
    print("announce scenario setup is finished")






################################################################################
#def get_data(**kwargs):
#    """custom algorithm to create a temporary game to retrieve true game data
#    NOTE: very similar to pysc2.bin.gen_actions.get_data()
#    """
#    run_config = run_configs.get()
#    with run_config.start(**kwargs) as controller:
#        m = maps.get("Sequencer")  # Arbitrary ladder map.
#        create = sc_pb.RequestCreateGame(local_map=sc_pb.LocalMap(
#            map_path=m.path, map_data=m.data(run_config)))
#        create.player_setup.add(type=sc_pb.Participant)
#        create.player_setup.add(type=sc_pb.Computer, race=sc_common.Random,
#                                difficulty=sc_pb.VeryEasy)
#        join = sc_pb.RequestJoinGame(race=sc_common.Random,
#                                     options=sc_pb.InterfaceOptions(raw=True))
#        controller.create_game(create)
#        controller.join_game(join)
#        return controller.data()


from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

from s2clientprotocol import sc2api_pb2 as sc_pb

from pysc2.lib import protocol
from pysc2.lib import remote_controller
from pysc2.lib.sc_process import FLAGS
from sc2common import types as t
from sc2gameLobby import gameConfig
from sc2gameLobby import gameConstants as c
from sc2maptool import selectMap
from sc2players import PlayerPreGame
#from sc2gamemgr import genericObservation as go

import os
import sys
import time

now = time.time


################################################################################
def run(debug=False):
    """PURPOSE: start a starcraft2 process using the defined the config parameters"""
    FLAGS(sys.argv)
    config = gameConfig.Config(
        version=None, # vers is None... unless a specific game version is desired
        themap=selectMap(), # pick a random map
        players=["defaulthuman", "blizzbot2_easy"],
    )
    createReq = sc_pb.RequestCreateGame( # used to advance to "Init Game" state, when hosting
        realtime    = config.realtime,
        disable_fog = config.fogDisabled,
        random_seed = int(now()), # a game is created using the current second timestamp as the seed
        local_map   = sc_pb.LocalMap(map_path=config.mapLocalPath,
                                     map_data=config.mapData) )
    joinRace = None
    for player in config.players:
        reqPlayer = createReq.player_setup.add() # add new player; get link to settings
        playerObj = PlayerPreGame(player)
        if playerObj.isComputer:
              reqPlayer.difficulty  = playerObj.difficulty.gameValue()
              pType                 = playerObj.type.type
        else: pType                 = c.PARTICIPANT
        reqPlayer.type              = t.PlayerControls(pType).gameValue()
        reqPlayer.race              = playerObj.selectedRace.gameValue()
        if not playerObj.isComputer:
            joinRace                = reqPlayer.race
    interface = sc_pb.InterfaceOptions()
    raw,score,feature,rendered = config.interfaces
    interface.raw   = raw   # whether raw data is reported in observations
    interface.score = score # whether score data is reported in observations
    interface.feature_layer.width = 24
    joinReq = sc_pb.RequestJoinGame(options=interface) # SC2APIProtocol.RequestJoinGame
    joinReq.race = joinRace # update joinGame request as necessary
    if debug: print("Starcraft2 game process is launching.")
    controller = None
    with config.launchApp() as controller:
      try:
        if debug: print("Starcraft2 application is live. (%s)"%(controller.status)) # status: launched
        controller.create_game(createReq)
        if debug: print("Starcraft2 is waiting for %d player(s) to join. (%s)"%(config.numAgents, controller.status)) # status: init_game
        playerID = controller.join_game(joinReq).player_id # SC2APIProtocol.RequestJoinGame
        print("[GET IN-GAME DATA] player#%d %s"%(playerID, config))
        return (controller.ping(), controller.data())
      except (protocol.ConnectionError, protocol.ProtocolError, remote_controller.RequestError) as e:
        if debug:   print("%s Connection to game closed (NOT a bug)%s%s"%(type(e), os.linesep, e))
        else:       print(   "Connection to game closed.")
      except KeyboardInterrupt:
        print("caught command to forcibly shutdown Starcraft2 host server.")
      finally:
        config.disable() # if the saved cfg file still exists, always remove it
        controller.quit() # force the sc2 application to close

